package simpledb;

import java.io.*;
import java.util.*;

/**
 * HeapFile is an implementation of a DbFile that stores a collection
 * of tuples in no particular order.  Tuples are stored on pages, each of
 * which is a fixed size, and the file is simply a collection of those
 * pages. HeapFile works closely with HeapPage.  The format of HeapPages
 * is described in the HeapPage constructor.
 *
 * @see simpledb.HeapPage#HeapPage
 * @author Sam Madden
 */
public class HeapFile implements DbFile {
    private final File f;

    int tableid = -1;
    /**
     * Constructor.
     * Creates a new heap file that stores pages in the specified buffer pool.
     *
     * @param f The file that stores the on-disk backing store for this DbFile.
     */
    public HeapFile(File f) {
        this.f = f;
        this.tableid = f.getAbsoluteFile().hashCode();
    }

    /**
     * Return a Java File corresponding to the data from this HeapFile on disk.
     */
    public File getFile() {
        return f;
    }

    /**
     * @return an ID uniquely identifying this HeapFile
     *  (Implementation note:  you will need to generate this tableid somewhere,
     *    ensure that each HeapFile has a "unique id," and that you always
     *    return the same value for a particular HeapFile.  The implementation we
     *    suggest you use could hash the absolute file name of the file underlying
     *    the heapfile, i.e. f.getAbsoluteFile().hashCode()
     *    )
     */
    public int id() {
        return tableid;
    }

    /**
     * Returns a Page from the file.
     */
    public Page readPage(PageId pid) throws NoSuchElementException {
        HeapPageId id = (HeapPageId) pid;
        BufferedInputStream bis = null;

        try {
            bis =  new BufferedInputStream(new FileInputStream(f));
            byte pageBuf[] = new byte[bytesPerPage()];
            if (bis.skip(id.pageno()*bytesPerPage()) != id.pageno()*bytesPerPage()) {
                throw new IllegalArgumentException("Unable to seek to correct place in heapfile");
            }
            int retval = bis.read(pageBuf, 0, bytesPerPage());
            if(retval == -1) {
                throw new IllegalArgumentException("Read past end of table");
            }
            if(retval < bytesPerPage()) {
                throw new IllegalArgumentException("Unable to read " + bytesPerPage() + " bytes from heapfile");
            }
            Debug.log(1, "HeapFile.readPage: read page %d", id.pageno());
            HeapPage p = new HeapPage(id, pageBuf);
            return p;
        } catch(IOException e) {
            throw new RuntimeException(e);
        } finally {
            // Close the file on success or error
            try {
                if (bis != null) bis.close();
            } catch (IOException ioe) {
                // Ignore failures closing the file
            }
        }
    }

    /**
     * Writes the given page to the appropriate location in the file.
     */
    public void writePage(Page page) throws IOException {
        // some code goes here
    }

    /**
     * Returns the number of pages in this HeapFile.
     */
    public int numPages() {
        return (int)(f.length() / bytesPerPage());
    }

    /**
     * Adds the specified tuple to the table under the specified TransactionId.
     *
     * @throws DbException
     * @throws IOException
     * @return An ArrayList contain the pages that were modified
     */
    public ArrayList<Page> addTuple(TransactionId tid, Tuple t)
        throws DbException, IOException, TransactionAbortedException {
        // some code goes here
        return null;
    }

    /**
     * Deletes the specified tuple from the table, under the specified
     * TransactionId.
     */
    public Page deleteTuple(TransactionId tid, Tuple t)
        throws DbException, TransactionAbortedException {
        // some code goes here
        return null;
    }

    /**
     * An iterator over all tuples on this file, over all pages.
     * Note that this iterator should use BufferPool.getPage(), rather than HeapFile.getPage()
     * to iterate through pages.
     */
    public DbFileIterator iterator(TransactionId tid) {
        return new HeapFileIterator(this, tid);
    }

    /**
     * Returns the number of bytes in the header of a page in the HeapFile.
     * Assumes sizeof(int) == 4.
     */
    private int headerBytes() {
        TupleDesc td = Database.getCatalog().getTupleDesc(tableid);
        int hb = (((BufferPool.PAGE_SIZE / td.getSize()) / 32) +1) * 4;
        return hb;
    }

    /**
     * @return the number of bytes on a page, including the number of bytes
     * in the header.
     */
    public int bytesPerPage() {
        return BufferPool.PAGE_SIZE + headerBytes();
    }
}

/**
 * Helper class that implements the Java Iterator for tuples on
 * a HeapFile
 */
class HeapFileIterator extends AbstractDbFileIterator {
    Iterator<Tuple> it = null;
    int curpgno = 0 ;

    TransactionId tid;
    HeapFile hf;

    public HeapFileIterator(HeapFile hf, TransactionId tid) {
        this.hf = hf;
        this.tid = tid;
    }

    public void open()
        throws DbException, TransactionAbortedException {
        curpgno = -1;
    }

    @Override
    protected Tuple readNext() throws TransactionAbortedException, DbException {
        if (it != null && !it.hasNext()) it = null;

        while (it == null && curpgno < hf.numPages()-1) {
            curpgno++;
            HeapPageId curpid = new HeapPageId(hf.id(), curpgno);
            HeapPage curp = (HeapPage) Database.getBufferPool().getPage(tid, curpid,
                                                               Permissions.READ_ONLY);
            it = curp.iterator();
            if (!it.hasNext()) it = null;
        }

        if (it == null) return null;
        return it.next();
    }

    public void rewind() throws DbException, TransactionAbortedException{
        close();
        open();
    }

    public void close() {
        super.close();
        it = null;
        curpgno = Integer.MAX_VALUE;
    }
}
